import django
from django.core.signals import setting_changed
from django.conf import settings as django_settings

DEFAULTS = {
    'DJANGO_VERSION': django.get_version()
}


class SettingNotFound(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class Settings(object):
    prefix = 'MIGRATION_TEST_'

    def __init__(self):
        self.reset_settings()
        setting_changed.connect(self.set_setting)

    def __getattr__(self, item):
        split_name = item.split(self.prefix)
        if len(split_name) > 1 and split_name[1] not in DEFAULTS:
            raise SettingNotFound(f'Can\'t find setting {item}.')
        value = self.get_setting(item)
        setattr(self, item, value)
        return value

    def get_setting(self, setting):
        assert setting.startswith(self.prefix), f'Unknown setting {setting}'
        return getattr(django_settings, setting, DEFAULTS[setting.split(self.prefix)[1]])

    def set_setting(self, setting, value, enter=False, **kwargs):
        if not setting.startswith(self.prefix):
            return
        setting = setting[len(self.prefix):]

        # ensure a valid app setting is being overridden
        if setting not in DEFAULTS:
            return

        # if exiting, delete value to repopulate
        if enter:
            setattr(self, setting, value)
        else:
            delattr(self, setting)

    def reset_settings(self):
        for setting, value in DEFAULTS.items():
            setattr(self, setting, value)



from django.apps import apps
from django.db import connection
from django.db.migrations.executor import MigrationExecutor
from django.test import TransactionTestCase

from dj_migration_test.config import Settings
from dj_migration_test.exceptions import AppNotFound


class MigrationTestCase(TransactionTestCase):
    """
    TestCase to write automated migration tests.

    To test migrations you need to add the attribute
        - migration_to: ('app_name', 'migration_name')-Tuple with the desired post-migration-state to perform tests on
        OPTIONAL
        - migration_from: List of ('app_name', 'migration_name')-Tuples containing the start state, including
        dependencies
        **you will need to add all dependencies by yourself**
    And implement the method `setUpDataBeforeMigration` which gives you the possibility to create database entries
    in state of the migration prior to `migration_to`
    """
    __settings = Settings()
    migrate_from = None
    migrate_to = None

    @property
    def app(self):
        try:
            return apps.get_containing_app_config(self.__module__).name.split('.')[-1]
        except AttributeError:
            raise AppNotFound()
        finally:
            return None

    def _perform_migration(self, migration):
        self._executor.migrate(migration)
        return self._executor.loader.project_state(migration)

    def _get_migration_states(self):
        try:
            migrate_to = self._executor.loader.get_migration(*self.migrate_to)
        except KeyError:
            raise AppNotFound()
        if not self.migrate_from:
            migrate_from = migrate_to.dependencies
        else:
            migrate_from = self.migrate_from if isinstance(self.migrate_from, list) else [self.migrate_from]
        return migrate_from, [self.migrate_to]

    def setUp(self):
        assert self.migrate_to, \
            f'TestCase \'{type(self).__name__}\' must define migrate_to property'

        self._executor = MigrationExecutor(connection)

        migrate_from, migrate_to = self._get_migration_states()

        # revert to origin state
        old_state = self._perform_migration(migrate_from)

        # prepare data for testing based on old state
        self.setUpDataBeforeMigration(self, old_state.apps)

        # reload graph
        self._executor.loader.build_graph()

        # Run the migration to perform tests on
        self.apps = self._perform_migration(migrate_to).apps

    @staticmethod
    def setUpDataBeforeMigration(test_instance, apps):  # noqa: N802
        """Create your testing data here.
        Use :attr apps: to work with pre-migration models.

        Example:
            test_model_class = apps.get_model('test_app', 'TestModelA')

            self.thread = test_model_class.objects.create(...)
        """
        raise NotImplementedError



import os
from setuptools import setup

with open('README', 'r') as fh:
    long_description = fh.read()

setup(
    name='dj-migration-test',
    packages=['dj_migration_test'],
    version=os.environ.get('TAG', 'v0.3.2'),
    license='MIT',
    description='Django migration test module',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Philipp Zettl',
    author_email='philipp.zettl@godesteem.de',
    url='https://github.com/philsupertramp/dj-migration-test',
    download_url=f'https://github.com/philsupertramp/dj-migration-test/archive/{os.environ.get("TAG", "v0.3.2")}.tar.gz',
    keywords=['Django', 'Testing', 'Database'],
    install_requires=[
        'django>=2.2'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        # '3 - Alpha', '4 - Beta' or '5 - Production/Stable'
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Bug Tracking',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
    ],
)
